#include "myopengl.h"
#include <QtOpenGL>

MyLine tempLine;
int drawTempLine = 0;

MyTriangle tempTriangle;
int drawTempTriangle = 0;
int triangleCount = 0;

const MyPoint *selectedPoint = NULL;

/* Дополнительные функции */

template <class T1, class T2>
inline float scalar(const T1 &a, const T2 &b) {
	return a.x() * b.x() + a.y() * b.y();
}

template <class T>
inline float length(const T &v) {
	return sqrtf(scalar(v, v));
}

template <class T>
inline T normalize(const T &v) {
	return v / length(v);
}

/* Реазизация класса рисования */

MyOpenGL::MyOpenGL()
{/*
	lines << MyLine(10, 10, 200, 300);
	lines << MyLine(700, 300, 20, 350);
	lines << MyLine(230, 400, 300, 350);
	lines << MyLine(350, 270, 400, 100);
*/

	// Точки полигона отсечения (выпуклый полигон)
	addToPolyClip(MyPoint(50, 280));
	addToPolyClip(MyPoint(150, 420));
	addToPolyClip(MyPoint(500, 500));
	addToPolyClip(MyPoint(700, 400));
	addToPolyClip(MyPoint(750, 70));
	addToPolyClip(MyPoint(400, 30));

	// Цветочек и солнышко
	int mas[][3][2] = {
{ { 344 , 129 } , { 385 , 94 } , { 355 , 164 } },
{ { 355 , 164 } , { 397 , 176 } , { 385 , 95 } },
{ { 385 , 95 } , { 423 , 126 } , { 398 , 178 } },
{ { 345 , 130 } , { 311 , 62 } , { 264 , 116 } },
{ { 355 , 164 } , { 272 , 165 } , { 313 , 222 } },
{ { 398 , 177 } , { 376 , 247 } , { 468 , 217 } },
{ { 424 , 124 } , { 486 , 185 } , { 465 , 36 } },
{ { 386 , 97 } , { 334 , 42 } , { 441 , 40 } },
{ { 374 , 169 } , { 319 , 304 } , { 331 , 321 } },
{ { 331 , 321 } , { 320 , 304 } , { 318 , 458 } },
{ { 318 , 458 } , { 309 , 508 } , { 321 , 508 } },
{ { 325 , 399 } , { 350 , 359 } , { 356 , 372 } },
{ { 354 , 365 } , { 398 , 281 } , { 422 , 354 } },
{ { 422 , 354 } , { 468 , 306 } , { 397 , 281 } },
{ { 397 , 281 } , { 469 , 306 } , { 482 , 281 } },
{ { 482 , 281 } , { 400 , 281 } , { 470 , 254 } },
{ { 319 , 368 } , { 308 , 334 } , { 296 , 333 } },
{ { 302 , 333 } , { 227 , 282 } , { 302 , 293 } },
{ { 302 , 293 } , { 228 , 283 } , { 271 , 215 } },
{ { 271 , 215 } , { 227 , 284 } , { 217 , 223 } },
{ { 97 , 504 } , { 309 , 507 } , { 256 , 501 } },
{ { 321 , 508 } , { 499 , 490 } , { 589 , 505 } },
{ { 681 , 39 } , { 663 , 74 } , { 735 , 84 } },
{ { 735 , 84 } , { 663 , 73 } , { 708 , 108 } },
{ { 663 , 73 } , { 546 , 51 } , { 553 , 73 } },
{ { 664 , 74 } , { 668 , 104 } , { 708 , 109 } },
{ { 669 , 104 } , { 574 , 167 } , { 591 , 176 } },
{ { 709 , 109 } , { 727 , 215 } , { 741 , 213 } },
{ { 682 , 40 } , { 658 , 2 } , { 673 , 0 } },
{ { 682 , 42 } , { 736 , 85 } , { 722 , 48 } },
{ { 722 , 48 } , { 750 , 2 } , { 758 , 3 } },
{ { 736 , 84 } , { 865 , 96 } , { 861 , 111 } }
	};

	// Формируем треугольники для рисования
	for (int i = 0; i < 32; i++) {
		tempTriangle.p1 = MyPoint(mas[i][0][0], mas[i][0][1]);
		tempTriangle.p2 = MyPoint(mas[i][1][0], mas[i][1][1]);
		tempTriangle.p3 = MyPoint(mas[i][2][0], mas[i][2][1]);
		triangles << tempTriangle;
	}
}

void MyOpenGL::initializeGL()
{
	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
	glPointSize(7.0f);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POINT_SMOOTH);
}

void MyOpenGL::resizeGL(int width, int height) {
	glViewport(0, 0, width, height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0, width, height, 0);
	glMatrixMode(GL_MODELVIEW);
}

void MyOpenGL::paintGL()
{
	glClear(GL_COLOR_BUFFER_BIT);

	glLineWidth(2.0f);

	// Треугольники, закрашенные

	glLineWidth(1.0f);
	glColor3f(0.0f, 0.8f, 0.0f);
	foreach (MyTriangle trig, triangles)
		fillTriangle(&trig.p1, &trig.p2, &trig.p3);

	glLineWidth(3.0f);
	glColor3f(0.0f, 0.4f, 0.0f);
	foreach (MyTriangle trig, triangles) {
		glBegin(GL_LINE_LOOP);
			glVertex2d(trig.p1.x(), trig.p1.y());
			glVertex2d(trig.p2.x(), trig.p2.y());
			glVertex2d(trig.p3.x(), trig.p3.y());
		glEnd();
	}

	// Треугольник для анимации рисования треугольника

	if (drawTempTriangle) {
		if (triangleCount > 1) {
			glLineWidth(0.5f);
			glColor3f(0.2f, 0.9f, 0.2f);
			fillTriangle(&tempTriangle.p1, &tempTriangle.p2, &tempTriangle.p3);
			glLineWidth(3.0f);
			glColor3f(0.2f, 0.5f, 0.2f);
			glBegin(GL_LINE_LOOP);
				glVertex2d(tempTriangle.p1.x(), tempTriangle.p1.y());
				glVertex2d(tempTriangle.p2.x(), tempTriangle.p2.y());
				glVertex2d(tempTriangle.p3.x(), tempTriangle.p3.y());
			glEnd();
	} else if (triangleCount > 0) {
			glLineWidth(3.0f);
			glColor3f(0.2f, 0.5f, 0.2f);
			glBegin(GL_LINES);
				glVertex2d(tempTriangle.p1.x(), tempTriangle.p1.y());
				glVertex2d(tempTriangle.p2.x(), tempTriangle.p2.y());
			glEnd();
		}
	}

	// Линии для рисования
	foreach (MyLine lineToClip, lines) {
		glColor3f(0.8f, 0.8f, 1.0f);
		glBegin(GL_LINES);
			glVertex2f(lineToClip.x1(), lineToClip.y1());
			glVertex2f(lineToClip.x2(), lineToClip.y2());
		glEnd();
		glColor3f(0.0f, 0.0f, 0.4f);
		drawClipLine(lineToClip);
	}

	// Линия для анимации рисования линии
	if (drawTempLine) {
		glColor3f(0.5f, 0.5f, 0.5f);
		glBegin(GL_LINES);
			glVertex2f(tempLine.x1(), tempLine.y1());
			glVertex2f(tempLine.x2(), tempLine.y2());
		glEnd();
	}

	// Отсекающий полигон (выпуклый)

	glLineWidth(3.0f);
	glColor3f(0.7f, 0.2f, 0.2f);
	glBegin(GL_LINE_LOOP);
	foreach (MyOrientedPoint orientedPoint, polyClip)
		glVertex2f(orientedPoint.point.x(), orientedPoint.point.y());
	glEnd();

	// Вершины отсекателя

	glColor3f(0.5f, 0.0f, 0.0f);
	glBegin(GL_POINTS);
	foreach (MyOrientedPoint orientedPoint, polyClip)
		glVertex2f(orientedPoint.point.x(), orientedPoint.point.y());
	glEnd();
}

void MyOpenGL::mousePressEvent(QMouseEvent *event)
{
	if (!drawTempLine && !drawTempTriangle) {
		if (event->buttons() & Qt::LeftButton)
		{
			tempLine.setPoints(event->pos(), event->pos());
			drawTempLine = 1;
//			qDebug() << tr("Левая кнопка мышки нажата");
			updateGL();
		}
	}
	if (event->buttons() & Qt::RightButton) {
		drawTempTriangle = 1;
		if (drawTempTriangle) {
			setMouseTracking(true);
			switch (triangleCount++) {
			case 0 : tempTriangle.p1 = tempTriangle.p2 = tempTriangle.p3 = event->pos(); break;
			case 1 : tempTriangle.p2 = tempTriangle.p3 = event->pos(); break;
			case 2 : tempTriangle.p3 = event->pos(); break;
			}
			if (triangleCount > 2) {
				setMouseTracking(false);
			}
			updateGL();
		}
//		qDebug() << tr("Правая кнопка мышки нажата");
	}
	QGLWidget::mousePressEvent(event);
}

void MyOpenGL::mouseReleaseEvent(QMouseEvent *event)
{
	//qDebug() << tr("Комбинация отпущеных кнопок мыши = %1").arg(event->buttons());
	if (drawTempLine) {
		if (event->button() == Qt::LeftButton) {
			if (tempLine.p1() != tempLine.p2())
				lines << tempLine;
			drawTempLine = 0;
//			qDebug() << tr("Левая кнопка мышки отпущена");
			updateGL();
		}
	}
	if (drawTempTriangle) {
		if (event->button() == Qt::RightButton) {
			if (triangleCount > 2) {
				triangleCount = 0;
				drawTempTriangle = 0;
				triangles << tempTriangle;
			}
//			qDebug() << tr("Правая кнопка мышки отпущена");
			updateGL();
		}
	}
	QGLWidget::mouseReleaseEvent(event);
}

void MyOpenGL::mouseMoveEvent(QMouseEvent *event)
{
	if (drawTempLine) {
		if (event->buttons() & Qt::LeftButton) {
			tempLine.setP2(event->pos());
			updateGL();
		}
	}
	if (drawTempTriangle) {
		switch (triangleCount) {
		case 0 : tempTriangle.p1 = tempTriangle.p2 = tempTriangle.p3 = event->pos(); break;
		case 1 : tempTriangle.p2 = tempTriangle.p3 = event->pos(); break;
		case 2 : tempTriangle.p3 = event->pos(); break;
		}
		updateGL();
	}
	QGLWidget::mouseMoveEvent(event);
}

/* Добавление точки в полигон отсечения */

void MyOpenGL::addToPolyClip(const MyPoint &point)
{
	MyOrientedPoint newPoint;
	newPoint.point = point;

	if (polyClip.isEmpty()) {
		newPoint.normal = QPointF(0.0f, 0.0f);
		polyClip.push_back(newPoint);
		return;
	}

	MyOrientedPoint nextPoint = polyClip.first();
	MyPoint vectorToNext = nextPoint.point - point;
	newPoint.normal = normalize( QPointF( -vectorToNext.y(), vectorToNext.x() ) );

	MyOrientedPoint prevPoint = polyClip.last();
	MyPoint vectorFromPrev = point - prevPoint.point;
	prevPoint.normal = normalize( QPointF( -vectorFromPrev.y(), vectorFromPrev.x() ) );

	polyClip.pop_back();
	polyClip.push_back(prevPoint);
	polyClip.push_back(newPoint);
}

/* Алгоритм Сайруса-Бека */

int MyOpenGL::chop(float *tIn, float *tOut, float numer, float denom)
{
	float tHit = numer / denom;
	if (denom < 0) {
		if (tHit > *tOut) return 0;
		if (tHit > *tIn) *tIn = tHit;
	} else if (denom > 0) {
		if (tHit < *tIn) return 0;
		if (tHit < *tOut) *tOut = tHit;
	}
	return 1;
}

int MyOpenGL::clipLine(const MyLine &lineToClip, QLineF *lineToDraw)
{
	*lineToDraw = lineToClip;
	if (polyClip.size() < 3)
		return 1;

	QPointF ab(lineToClip.p2() - lineToClip.p1());
	float tIn = 0, tOut = 1;

	foreach (MyOrientedPoint p, polyClip) {
		if ( !chop(&tIn, &tOut, scalar(p.point - lineToClip.p1(), p.normal), scalar(ab, p.normal)) )
			return 0;
	}
	*lineToDraw = QLineF(lineToClip.p1() + tIn * ab, lineToClip.p1() + tOut * ab);
	return 1;
}

void MyOpenGL::drawClipLine(const MyLine &lineToClip)
{
	QLineF clipedLine = lineToClip;
	if (clipLine(lineToClip, &clipedLine)) {
		glBegin(GL_LINES);
			glVertex2d(clipedLine.x1(), clipedLine.y1());
			glVertex2d(clipedLine.x2(), clipedLine.y2());
		glEnd();
	}
}

/* Закраска  треугольника */

void MyOpenGL::fillTriangle(const MyPoint *max, const MyPoint *mid, const MyPoint *min)
{
	// поиск верхней, средней и нижней точки
	if (max->y() <= mid->y()) qSwap(max, mid);
	if (max->y() <= min->y()) qSwap(max, min);
	if (mid->y() <= min->y()) qSwap(mid, min);

	// вычисление угла наклона (если при вычислении деление на ноль, то закрашивание не нужно)
	float dx01 = 0, dx02 = 0, dx12 = 0;
	int temp;

	if ((temp = max->y() - mid->y()) != 0) dx12 = (max->x() - mid->x()) / (float)temp;
	if ((temp = max->y() - min->y()) != 0) dx02 = (max->x() - min->x()) / (float)temp;
	if ((temp = mid->y() - min->y()) != 0) dx01 = (mid->x() - min->x()) / (float)temp;

	// два цикла закраски
	float x1, x2 = x1 = min->x();
	int y = min->y();

	// если min и mid на одной высоте, то x1 и x2 должны различаться
	if (temp == 0) { x1 = mid->x(); }

	while (y < mid->y()) {
		drawClipLine(MyLine(MyPoint(x1, y), MyPoint(x2, y)));
		x1 += dx01;
		x2 += dx02;
		y++;
	}
	while (y < max->y()) {
		drawClipLine(MyLine(MyPoint(x1, y), MyPoint(x2, y)));
		x1 += dx12;
		x2 += dx02;
		y++;
	}
}
